import { isObject, isString, isFunction, isElement, isUndefined, isBoolean, noop, isTrue } from './assertions';
import Observer from './observer';

const internals = {};

internals.components = [];

internals.$state = Symbol('state');

internals.getTag = object => object && object.tagName;

internals.defaultUpdater = (element, newProps, update) => {

  update(newProps);

};

internals.parseOptions = options => {

  isObject(options, 'options');
  isString(options.tagName, 'tagName');

  options.update && isFunction(options.update, 'update');
  options.render && isFunction(options.render, 'render');
  options.connect && isFunction(options.connect, 'connect');
  options.disconnect && isFunction(options.disconnect, 'disconnect');

  options.role && isString(options.role, 'role');
  options.tabIndex && isString(options.tabIndex, 'tabIndex');
  options.initialState && isObject(options.initialState, 'initialState');

  const attributes = [];

  if (options.attributes !== void 0) {

    isObject(options.attributes, 'attributes');

    Object.keys(options.attributes).reduce((attributes, attributeName) => {

      isString(attributeName, 'attributeName');

      const { defaultValue = '' } = isObject(options.attributes[attributeName], 'attribute');

      const attribute = {
        attributeName,
        defaultValue
      };

      attributes.push(Object.freeze(attribute));

      return attributes;
    }, attributes);
  }
  options.attributes = attributes;

  isObject(options.observer, 'observer');
  isTrue(options.observer.isPwetObserver, 'isPwetObserver');

  return options;
};

internals.searchElements = (tagName, element = document.body) => Array.from(element.getElementsByTagName(tagName));

internals.Component = module.exports = (options = {}) => {

  const {
    observer,
    tagName,
    role,
    tabIndex,
    initialState = {},
    attributes,
    update = internals.defaultUpdater,
    render = noop,
    connect:_connect = noop,
    disconnect:_disconnect = noop,
    ...properties
  } = internals.parseOptions(options);

  if (internals.components.map(internals.getTag).includes(tagName))
    throw new Error(`A "${tagName}" component is already registered`);


  const component = {
    isPwetComponent: true,
    tagName,
    attributes,
    render,
    update,
    connect(element) {

      console.log('Component.connect()', tagName,  element);

      // if (!isUndefined(element.__component))
      //   return;
      isUndefined(element.__component, 'That element is already connected', true);


      if (role !== void 0)
        element.setAttribute('role', role);

      if (tabIndex !== void 0)
        element.setAttribute('tabIndex', tabIndex);

      Object.defineProperty(element, 'state', {
        get() {
          return element[internals.$state];
        },
        set(newState) {

          console.log('setState()', newState);

          component.update(element, newState, (state, render = true) => {


            element[internals.$state] = state;

            if (render)
              component.render(element, state, Object.keys(properties).reduce((before, key) => {

                let value = properties[key];

                if (isFunction(value))
                  value = value.bind(null, element);

                return Object.assign(before, { [key]: value })
              }, {}));

          });
        }
      });

      element.state = attributes.reduce((state, attribute) => {

        const { attributeName, defaultValue } = attribute;

        Object.defineProperty(element, attributeName, {
          get() {
            return element.state[attributeName];
          },
          set(newValue) {

            console.log('setAttribute()', newValue);

            element.state = Object.assign(element.state, {
              [attributeName]: newValue
            });
          }
        });

        return Object.assign(state, {
          [attributeName]: element.getAttribute(attributeName) || defaultValue
        });

      }, Object.assign({}, initialState));


      if (attributes.length > 0) {

        element.__observer = Observer(element, { attributes: true });

        const attributesHasChanged = ({ element, name, oldValue }) => {

          if (!(name in attributes))
            return;

          element.state = Object.assign(element.state, {
            [name]: element.getAttribute(name)
          });
        };

        attributesHasChanged.tagName = component.tagName;

        element.__observer.attributesHasChanged.push(attributesHasChanged);
        element.__observer.connect();
      }

      element.__component = component;

      // if (_elements.length < 1)
      //   observer.connect();

      _connect(element);

      element.dispatchEvent(new CustomEvent("attached", {
        bubbles: true,
        cancelable: false
      }));

      // _elements.push(element);
    },
    disconnect(element) {

      console.error('Component.disconnect()', tagName, element);

      // const index = _elements.indexOf(element);
      //
      // if (!~index)
      //   return;

      _disconnect(element);

      // _elements.splice(index, 1);

      // if (_elements.length < 1)
      //   observer.disconnect();
    }
  };

  const wasAttached = element => {

    console.error('attached1', element.tagName.toLowerCase(), component.tagName);

    if (element.tagName.toLowerCase() === component.tagName) {

      component.connect(element);

      return;
    }
    console.error('attached2', element.tagName.toLowerCase(), component.tagName);

    internals.searchElements(tagName, element).forEach(component.connect);

  };


  const wasDetached = element => {

    console.error('detached1', element.tagName.toLowerCase(), component.tagName);

    if (element.tagName.toLowerCase() === component.tagName) {

      component.disconnect(element);

      return;
    }
    console.error('detached2', element.tagName.toLowerCase(), component.tagName);

    internals.searchElements(tagName, element).forEach(component.disconnect);

  };

  wasAttached.tagName = wasDetached.tagName = component.tagName;

  observer.wasAttached.push(wasAttached);
  // observer.wasDetached.push(wasDetached);

  internals.searchElements(component.tagName).forEach(component.connect);

  internals.components.push(component);

  return Object.freeze(component);
};
